local MapDims = LibStub("LibMapData-1.0")

local GetPlayerMapPosition = GetPlayerMapPosition
local SetMapToCurrentZone = SetMapToCurrentZone
local GetCurrentMapDungeonLevel = GetCurrentMapDungeonLevel
local GetNumRaidMembers = GetNumRaidMembers
local GetNumPartyMembers = GetNumPartyMembers
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsConnected = UnitIsConnected
local UnitIsVisible = UnitIsVisible
local UnitGUID = UnitGUID
local UnitIsUnit = UnitIsUnit
local WorldMapFrame = WorldMapFrame
local sort = sort
local tostring = tostring

local unitRaidIDs = {}
for i = 1,MAX_RAID_MEMBERS do
	unitRaidIDs[i] = format("raid%i", i)
end
local unitPartyIDs = {}
unitPartyIDs[1] = "player"
for i = 1,MAX_PARTY_MEMBERS do
	unitPartyIDs[i+1] = format("party%i", i)
end

local MySort = function(a,b)
	if a.isDeadOrOffline ~= b.isDeadOrOffline then
		return a.isDeadOrOffline < b.isDeadOrOffline
	else
		if a.isInRange ~= b.isInRange then
			return a.isInRange > b.isInRange
		else
			if a[0] ~= b[0] then
				return a[0] > b[0]
			else
				if a.healthDeficit ~= b.healthDeficit then
					return a.healthDeficit > b.healthDeficit
				else
					return a.unit < b.unit
				end
			end
		end
	end
end

local GridStatus = Grid:GetModule("GridStatus")
local GridRoster = Grid:GetModule("GridRoster")

local GridStatusBestCluster = GridStatus:NewModule("GridStatusBestCluster", "AceTimer-3.0")
GridStatusBestCluster.menuName = "GridStatusBestCluster"
GridStatusBestCluster.options = false
GridStatusBestCluster.data = {}
GridStatusBestCluster.data.distances = {}
GridStatusBestCluster.data.sortedList = {}
GridStatusBestCluster.data.colorsToDisplay = {}

local distances = GridStatusBestCluster.data.distances
local sortedList = GridStatusBestCluster.data.sortedList
local colorsToDisplay = GridStatusBestCluster.data.colorsToDisplay
local settings

GridStatusBestCluster.defaultDB = {
	alert_bestcluster = {
		text = "Best Cluster",
		enable = false,
		testrange = 7,
		frequency = 3,
		numToFind = 3,
		color = { r = 1.0, g = 0.0, b = 0.0, a = 1.0 },
		color2 = { r = 1.0, g = 0.0, b = 0.0, a = 0.3 },
		usedeficits = false,
		priority = 90,
		range = false,
		smWeight = 0
	},
}

local bestcluster_options = {
	["testrange"] = {
		type = "range",
		name = "Test range",
		desc = "Maximum range at which to consider two raid members \"clustered\" with eachother.",
		get =   function () 
                    return GridStatusBestCluster.db.profile.alert_bestcluster.testrange 
                end,
		set =   function (_, v)
                    GridStatusBestCluster.db.profile.alert_bestcluster.testrange = v
					GridStatusBestCluster.data.testRange = v^2
                end,
		min = 1,
		max = 40,
		step = 1,
		isPercent = false,
		order = -1,
	},
	["frequency"] = {
		type = "range",
		name = "Refresh frequency",
		desc = "Seconds between status refreshes",
		get =   function () 
                    return GridStatusBestCluster.db.profile.alert_bestcluster.frequency 
                end,
		set =   function (_, v)
                    GridStatusBestCluster.db.profile.alert_bestcluster.frequency = v
                    GridStatusBestCluster:CancelTimer(GridStatusBestCluster.UpdateAllUnitsTimer, true)
                    if GridStatusBestCluster.db.profile.alert_bestcluster.enable then
						GridStatusBestCluster.UpdateAllUnitsTimer = GridStatusBestCluster:ScheduleRepeatingTimer("UpdateAllUnits", GridStatusBestCluster.db.profile.alert_bestcluster.frequency)
					end
                end,
		min = 0,
		max = 5,
		step = 0.1,
		isPercent = false,
		order = -1,
	},
	["numToFind"] = {
		type = "range",
		name = "Number of targets to find",
		desc = " ",
		max = 25,
		min = 1,
		step = 1,
		get =   function ()
                    return GridStatusBestCluster.db.profile.alert_bestcluster.numToFind
                end,
		set =   function (_, v)
                    GridStatusBestCluster.db.profile.alert_bestcluster.numToFind = v
                    GridStatusBestCluster:RedetermineColors()
                end,
   order = -1
	},
	["color"] = {
		type = "color",
		name = "Color 1",
		desc = "Best of the best color.",
		hasAlpha = true,
		get = function ()
			local color = GridStatusBestCluster.db.profile.alert_bestcluster.color
			return color.r, color.g, color.b, color.a
		end,
		set = function (_, r, g, b, a)
			local color = GridStatusBestCluster.db.profile.alert_bestcluster.color
			color.r = r
			color.g = g
			color.b = b
			color.a = a or 1
			GridStatusBestCluster:RedetermineColors()
		end,
	},
	["color2"] = {
		type = "color",
		name = "Color 2",
		desc = "Worst of the best color.",
		hasAlpha = true,
		get = function ()
			local color = GridStatusBestCluster.db.profile.alert_bestcluster.color2
			return color.r, color.g, color.b, color.a
		end,
		set = function (_, r, g, b, a)
			local color = GridStatusBestCluster.db.profile.alert_bestcluster.color2
			color.r = r
			color.g = g
			color.b = b
			color.a = a or 1
			GridStatusBestCluster:RedetermineColors()
		end,
	},
	["usedeficits"] = {
		type = "toggle",
		name = "Use Health Deficits",
		desc = "Determine best target by total health deficit of surrounding raid members instead of total number of surrounding raid members.",
		get = function () return GridStatusBestCluster.db.profile.alert_bestcluster.usedeficits end,
		set = function (_, v)
			GridStatusBestCluster.db.profile.alert_bestcluster.usedeficits = v
		end,
	},
	["smWeight"] = {
	type = "input",
	name = "Swiftmendable Bonus",
	desc = "The number here will be added to the number of targets found surrounding a raid member (or if Use Deficits is enabled, to the total health deficit of surrounding raid members). This gives a method of favoring Swiftmendable targets. You must have auras for Regrowth and Rejuvenation enabled in Grid to detect Swiftmendable targets.",
	get = function () return GridStatusBestCluster.db.profile.alert_bestcluster.smWeight end,
	set = function (_, v)
		if type(tonumber(v)) == "number" then
			GridStatusBestCluster.db.profile.alert_bestcluster.smWeight = v
		end
	end,
	},
}

function GridStatusBestCluster:RedetermineColors()
	wipe(colorsToDisplay)
	local color1 = self.db.profile.alert_bestcluster.color
	local color2 = self.db.profile.alert_bestcluster.color2
	if settings.numToFind >= 2 then
		for i = 1,settings.numToFind do
			local spread = {}
			spread.r = color2.r - color1.r
			spread.g = color2.g - color1.g
			spread.b = color2.b - color1.b
			spread.a = color2.a - color1.a
			local percDone = (i - 1) / (settings.numToFind - 1)
			colorsToDisplay[i] = {}
			colorsToDisplay[i].r = color1.r + percDone * spread.r
			colorsToDisplay[i].g = color1.g + percDone * spread.g
			colorsToDisplay[i].b = color1.b + percDone * spread.b
			colorsToDisplay[i].a = color1.a + percDone * spread.a
		end
	else
		colorsToDisplay[1] = {}
		colorsToDisplay[1].r = color1.r
		colorsToDisplay[1].g = color1.g
		colorsToDisplay[1].b = color1.b
		colorsToDisplay[1].a = color1.a
	end
end

function GridStatusBestCluster:OnInitialize()
	self.super.OnInitialize(self)
	self:RegisterStatus("alert_bestcluster", "Best Cluster", bestcluster_options, true)
	settings = self.db.profile.alert_bestcluster
	self:RedetermineColors()
end

function GridStatusBestCluster:OnEnable()
	self.super.OnEnable(self)
end

function GridStatusBestCluster:OnStatusEnable(status)
	self:RegisterMessage("Grid_RosterUpdated")
	self:RegisterMessage("Grid_UnitOffline")
	self.data.testRange = settings.testrange^2
	self:CancelTimer(self.UpdateAllUnitsTimer, true)
	self.UpdateAllUnitsTimer = self:ScheduleRepeatingTimer("UpdateAllUnits", self.db.profile.alert_bestcluster.frequency)
	self:Grid_RosterUpdated()
	MapDims.RegisterCallback(self,"MapChanged")
end

function GridStatusBestCluster:OnStatusDisable(status)
	self.core:SendStatusLostAllUnits(status)
	self:UnregisterAllEvents()
	self:UnregisterAllMessages()
	MapDims.UnregisterAllCallbacks(self)
	self:CancelTimer(self.UpdateAllUnitsTimer, true)
end

function GridStatusBestCluster:Reset()
	self.super.Reset(self)
	self.core:SendStatusLostAllUnits(status)
	self:UnregisterStatus("alert_bestcluster")
	self:RegisterStatus("alert_bestcluster", "BestCluster", nil, true)
	self:UpdateAllUnits()
end

function GridStatusBestCluster:Grid_UnitOffline(event, guid)
	self.core:SendStatusLost(guid, "alert_bestcluster")
end

function GridStatusBestCluster:MapChanged(event,map,level,w,h)
	self.data.dimsX, self.data.dimsY = w,h
	if not self.data.dimsX or self.data.dimsX == 0 then 
		SetMapToCurrentZone()
		local map = GetMapInfo()
		local level = GetCurrentMapDungeonLevel()
		self.data.dimsX,self.data.dimsY = MapDims:MapArea(map,level)
	end
	self:UpdateAllUnits()
end

function GridStatusBestCluster:Grid_RosterUpdated()
	if (not self) or (not self.db.profile.alert_bestcluster.enable) then return end
	self.data.numRaiders = GetNumRaidMembers()
	self.data.numPartyMembers = GetNumPartyMembers() + 1
	wipe(distances)
	if self.data.numRaiders > 0 then
		for i = 1,self.data.numRaiders do
			distances[i] = distances[i] or {}
			distances[i].GUID = UnitGUID(unitRaidIDs[i])
		end
	else
		for i = 1,self.data.numPartyMembers do
			distances[i] = distances[i] or {}
			distances[i].GUID =  UnitGUID(unitPartyIDs[i])
		end
	end
	self:CancelTimer(self.UpdateAllUnitsTimer, true)
	if self.data.numRaiders > 1 or self.data.numPartyMembers > 1 then
		self.UpdateAllUnitsTimer = self:ScheduleRepeatingTimer("UpdateAllUnits", self.db.profile.alert_bestcluster.frequency)
	else
		self.core:SendStatusLostAllUnits("alert_bestcluster")
	end
end

function GridStatusBestCluster:UpdateAllUnits()
	local numInterested
	if self.data.numRaiders > 0 then
		numInterested = self.data.numRaiders
	else
		numInterested = self.data.numPartyMembers
	end
	if numInterested > 1 then
		self:GetBestTargets()
		local goto = numInterested > settings.numToFind and settings.numToFind or numInterested
		for guid, unitid in GridRoster:IterateRoster() do
			for i = 1,goto do
				if sortedList[i] and UnitIsUnit(sortedList[i].unit, unitid) then
					if sortedList[i].isDeadOrOffline == 0 then
						local amt = sortedList[i][0] - settings.smWeight * sortedList[i].isSwiftmendable
						self.core:SendStatusGained(guid, "alert_bestcluster",
										settings.priority,
										(settings.range and 40),
										colorsToDisplay[i],
										amt > 0 and tostring(amt) or nil
						)
						break
					end
				end
				if i == goto then
					self.core:SendStatusLost(guid, "alert_bestcluster")
				end
			end
		end
	else
		self.core:SendStatusLostAllUnits("alert_bestcluster")
	end
end

function GridStatusBestCluster:GetBestTargets()
	--UpdateRaidLocations
	if not self.data.dimsX then
		SetMapToCurrentZone()
		local map = GetMapInfo()
		local level = GetCurrentMapDungeonLevel()
		self:MapChanged("MapChanged", map, level, MapDims:MapArea(map,level))
		if (not self.data.dimsX) or (self.data.dimxX == 0) then
			--print("Error: No map loaded.")
			return
		end
	end
	local numInterested, unitIDsInterested
	if self.data.numRaiders > 0 then
		numInterested = self.data.numRaiders
		unitIDsInterested = unitRaidIDs
	else
		numInterested = self.data.numPartyMembers
		unitIDsInterested = unitPartyIDs
	end
	for i = 1,numInterested do
		local unit = unitIDsInterested[i]
		local GUID = distances[i].GUID
		local x,y
		distances[i] = distances[i] or {}
		distances[i].isSwiftmendable = (GridStatus:GetCachedStatus(GUID, "buff_Rejuvenation") or GridStatus:GetCachedStatus(GUID, "buff_Regrowth")) and 1 or 0
		if not settings.range then
			distances[i].isInRange = 1
		else
			distances[i].isInRange = GridStatus:GetCachedStatus(GUID, "alert_range") and 0 or 1
		end
		if settings.usedeficits then
			distances[i].healthDeficit = UnitHealthMax(unit) - UnitHealth(unit)
		else
			distances[i].healthDeficit = 0
		end
		if UnitIsDeadOrGhost(unit) or (not UnitIsConnected(unit)) or (not UnitIsVisible(unit)) then
			distances[i][0] = 0
			distances[i].isDeadOrOffline = 1
			x,y = 0,0
		else
			distances[i][0] = settings.usedeficits and distances[i].healthDeficit or 1
			distances[i][0] = distances[i][0] + settings.smWeight * distances[i].isSwiftmendable
			distances[i].isDeadOrOffline = 0
			x,y = GetPlayerMapPosition(unit)
			if x <= 0 and y <= 0 then
				if not WorldMapFrame:IsShown() then
					SetMapToCurrentZone()
					x,y = GetPlayerMapPosition(unit)
				end
			end
		end
		distances[i].location = distances[i].location or {x = 0, y = 0}
		distances[i].oldLocation = distances[i].oldLocation or {x = 0, y = 0}
		distances[i].oldLocation.x, distances[i].oldLocation.y = distances[i].location.x, distances[i].location.y
		distances[i].location.x, distances[i].location.y = x * self.data.dimsX, y * self.data.dimsY
		distances[i].unit = unit
	end
	
	--UpdateDistances
	for i = 1,numInterested-1 do
		if distances[i].isDeadOrOffline == 0 then
			local x1,y1 = distances[i].location.x, distances[i].location.y
			for j = i+1,numInterested do
				if distances[j].isDeadOrOffline == 0 then
					local distance
					if true or (distances[i].location.x ~= distances[i].oldLocation.x)
					or (distances[j].location.x ~= distances[j].oldLocation.x)
					or (distances[i].location.y ~= distances[i].oldLocation.y)
					or (distances[j].location.y ~= distances[j].oldLocation.y) then
						local x2,y2 = distances[j].location.x, distances[j].location.y
						if x1 and x2 then
							local dx = x2 - x1
							local dy = y2 - y1
							distance = dx*dx + dy*dy
						else
							distance = nil
						end
						distances[j] = distances[j] or {}
						distances[i][j] = distance
						distances[j][i] = distance
					else
						distance = distances[i][j]
					end
					if distance <= self.data.testRange then
						if settings.usedeficits then
							distances[i][0] = distances[i][0] + distances[j].healthDeficit
							distances[j][0] = distances[j][0] + distances[i].healthDeficit
						else
							distances[i][0] = distances[i][0] + 1
							distances[j][0] = distances[j][0] + 1
						end
					end
				end
			end
		end
	end
	
	for i = 1,#distances do
		sortedList[i] = distances[i]
	end
	for i = #distances+1, #sortedList do
		sortedList[i] = nil
	end
	
	sort(sortedList, MySort)
end