local Type, Version = "UIF-StatusBar", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

local LSM = LibStub("LibSharedMedia-3.0")

local pairs = pairs

local CreateFrame, UIParent = CreateFrame, UIParent

local StatusbarHeight = 30
local StatusbarPadding = 2

local NormalColor = {1, 1, 1, 1}
local HoverColor = {1, 1, 0.5, 1}
local SelectedColor = {0.5, 1, 0.5, 1}

local StatusbarFrames = {}

local function StatusbarFrame_OnEnter(this)
	this:SetStatusBarColor(unpack(HoverColor))
end

local function StatusbarFrame_OnLeave(this)
	if this.checked then
		this:SetStatusBarColor(unpack(SelectedColor))
	else
		this:SetStatusBarColor(unpack(NormalColor))
	end
end

local function StatusbarFrame_OnClick(this, button)
	local self = this.obj
	for i, frame in pairs(self.StatusbarFrames) do
		if frame ~= this then	
			frame.checked = false
			frame:SetStatusBarColor(unpack(NormalColor))
		end
	end
	
	this:SetStatusBarColor(unpack(SelectedColor))
	this.checked = true
	
	self:Fire("OnValueChanged", this.statusbar)
	self:SetValue(this.statusbar)
	
	if button == "RightButton" then
		AceGUI:ClearFocus()
	end
end

local function CreateStatusbarFrame(self, statusbar)
	local texture = LSM:Fetch("statusbar", statusbar)	
	self.pullout.test:SetTexture(texture)
	if self.pullout.test:GetTexture() then
		local frame = table.remove(StatusbarFrames)
		if not frame then
			frame = CreateFrame("StatusBar")
			frame:EnableMouse(true)
			frame:SetScript("OnEnter", StatusbarFrame_OnEnter)
			frame:SetScript("OnLeave", StatusbarFrame_OnLeave)
			frame:SetScript("OnMouseDown", StatusbarFrame_OnClick)
			
			local t1 = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			t1:SetPoint("LEFT", 5, -2)
			t1:SetPoint("RIGHT")			
			t1:SetJustifyH("LEFT")
			
			frame.t1 = t1
		end
		
		frame.t1:SetText(statusbar)
		frame:SetStatusBarTexture(texture)
		frame.obj = self
		frame.statusbar = statusbar
		
		return frame
	end
end

local sortedBars = {}
local function StatusbarFrame_Show(this)
	local self = this.obj
	
	if self.pullout:IsShown() then
		AceGUI:ClearFocus()
	else
		self.pullout:SetPoint("BOTTOMRIGHT", self.pane, "TOPRIGHT")
		self.pullout:SetPoint("BOTTOMLEFT", self.pane, "TOPLEFT")
		self.pullout:SetHeight(200)
		
		wipe(sortedBars)
		for i, v in pairs(LSM:List(LSM.MediaType.STATUSBAR)) do
			tinsert(sortedBars, v)
		end
		table.sort(sortedBars, function(a, b) return string.upper(a) < string.upper(b) end)
		local count = 0
		local selectedIndex
		for i, v in pairs(sortedBars) do
			local f = CreateStatusbarFrame(self, v)
			if f then				
				count = count + 1
				f:SetParent(self.scrollchild)
				f:SetHeight(StatusbarHeight)
				if count == 1 then
					f:SetPoint("TOPLEFT", StatusbarPadding, 0)
					f:SetPoint("TOPRIGHT", -StatusbarPadding, 0)
				else
					f:SetPoint("TOPLEFT", self.StatusbarFrames[count - 1], "BOTTOMLEFT", 0, -StatusbarPadding)
					f:SetPoint("TOPRIGHT", self.StatusbarFrames[count - 1], "BOTTOMRIGHT", 0, -StatusbarPadding)
				end
				
				if f.statusbar == self.value then
					f:SetStatusBarColor(unpack(SelectedColor))
					selectedIndex = count - 1					
				else
					f:SetStatusBarColor(unpack(NormalColor))
				end
				
				tinsert(self.StatusbarFrames, f)
			end
		end
		
		self.scrollchild:SetHeight(ceil(#self.StatusbarFrames) * (StatusbarHeight + StatusbarPadding) - StatusbarPadding)
		if selectedIndex then
			local v = (selectedIndex - 1) * (StatusbarHeight + StatusbarPadding) - StatusbarPadding
			self.scrollbar:SetValue(v)
		end
		
		self.pullout:SetFrameStrata("FULLSCREEN_DIALOG")
		self.pullout:SetFrameLevel(this:GetFrameLevel() + 10)
		self.pullout:Show()
		AceGUI:SetFocus(self)
	end
end

--[[-----------------------------------------------------------------------------
Scripts
-------------------------------------------------------------------------------]]
local function Button_OnEnter(self)
	self.texture:SetTexture(0.2, 0.2, 0.6, 1)
end

local function Button_OnLeave(self)
	self.texture:SetTexture(0, 0.5, 0.5, 1)
end

local function Close_Pullout(self)
	for i, frame in pairs(self.StatusbarFrames) do
		frame:ClearAllPoints()
		frame.even = nil
		frame.checked = nil
		table.insert(StatusbarFrames, frame)
	end
	wipe(self.StatusbarFrames)
	self.pullout:Hide()
end

local function Scrollbar_OnValueChanged(this, value)
	this.obj.scrollframe:SetVerticalScroll(value)
end

local function ScrollFrame_OnSizeChanged(this)
	local self = this.obj
	self.scrollchild:SetWidth(self.scrollframe:GetWidth() - 20)
	
	local ScrollMax = max(self.scrollchild:GetHeight() - self.scrollframe:GetHeight(), 0)
	self.scrollbar:SetMinMaxValues(0, ScrollMax)
end

local function ScrollFrame_OnMouseWheel(this, value)
	local self = this.obj
	local min, max = self.scrollbar:GetMinMaxValues()
	local current = self.scrollbar:GetValue()
	if IsShiftKeyDown() and value > 0 then
	   self.scrollbar:SetValue(current - (StatusbarHeight - StatusbarPadding) * 2)
	elseif IsShiftKeyDown() and value < 0 then
	   self.scrollbar:SetValue(current + (StatusbarHeight + StatusbarPadding) * 2)
	elseif value < 0 and current < max then
	   self.scrollbar:SetValue(current + StatusbarHeight + StatusbarPadding)
	elseif value > 0 and current > 1 then
	   self.scrollbar:SetValue(current - StatusbarHeight - StatusbarPadding)
	end
end


--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
methods = {
	["OnAcquire"] = function(self)
		self:SetWidth(300)
		self:SetHeight(50)
	end,
	
	["OnRelease"] = function(self)
		Close_Pullout(self)
	end,
	
	["ClearFocus"] = function(self)
		Close_Pullout(self)
	end,
	
	["SetFont"] = function(self, font, size, flags)
		local f, s, fl = GameFontNormal:GetFont()
		self.title:SetFont(font or f, size or s, flags or fl)
		self.statusbar.Name:SetFont(font or f, s, fl)
	end,
	
	["SetFontColor"] = function(self, r, g, b, a)
		if type(r) == "table" then
			r, g, b, a = unpack(r)
		end
		self.title:SetTextColor(r, g, b, a or 1)		
	end,
	
	["SetBorderColor"] = function(self, r, g, b, a)
		if type(r) == "table" then
			r, g, b, a = unpack(r)
		end
		self.pane:SetBackdropBorderColor(r, g, b, a or 1)
	end,
	
	["SetValue"] = function(self, statusbar)
		self.statusbar:SetStatusBarTexture(LSM:Fetch("statusbar", statusbar))
		self.statusbar.Name:SetText(statusbar)
		self.value = statusbar
	end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]

local Backdrop  = {
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = GetUIFPath().."\\textures\\border3.tga",
	tile = true, tileSize = 16, edgeSize = 16,
	insets = { left = 2, right = 2, top = 2, bottom = 2 }
}

local function Constructor()	
	local count = AceGUI:GetNextWidgetNum(Type)
	local frame = CreateFrame("Frame", Type..count, UIParent)	
	
	local title = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	title:SetPoint("TOPLEFT")
	title:SetText("Statusbar")
	
	local pane = CreateFrame("Frame", nil, frame)
	pane:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -1)
	pane:SetPoint("BOTTOMRIGHT")
	pane:SetBackdrop(Backdrop)
	pane:SetBackdropColor(0, 0, 0, 1)
	pane:SetBackdropBorderColor(0.4, 0.4, 0.4)
	
	local button = CreateFrame("Frame", nil, pane)
	button:SetPoint("TOPRIGHT", -3, -3)
	button:SetPoint("BOTTOM", 0, 3)
	button:SetWidth(10)
	button:SetScript("OnMouseDown", StatusbarFrame_Show)
	button:SetScript("OnEnter", Button_OnEnter)
	button:SetScript("OnLeave", Button_OnLeave)
	local t = button:CreateTexture(nil, "ARTWROK")
	t:SetTexture(0, 0.5, 0.5, 1)
	t:SetAllPoints()
	button.texture = t
	
	local statusbar = CreateFrame("StatusBar", nil, pane)
	statusbar:SetPoint("TOPLEFT", 3, -3)
	statusbar:SetPoint("BOTTOMRIGHT", button, "BOTTOMLEFT", -2, 0)
	statusbar:SetStatusBarTexture(LSM:Fetch("statusbar", LSM:List(LSM.MediaType.STATUSBAR)[1]))
	
	local name = statusbar:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	name:SetPoint("LEFT", 5, 0)
	statusbar.Name = name
	name:SetText(LSM:List(LSM.MediaType.STATUSBAR)[1])
	
	local pullout = CreateFrame("Frame", nil, UIParent)
	Backdrop.edgeFile = GetUIFPath().."\\textures\\border1.tga"
	pullout:SetBackdrop(Backdrop)
	pullout:SetBackdropColor(0, 0, 0, 1)
	pullout:SetBackdropBorderColor(0.4, 0.4, 0.4, 1)
	pullout:EnableMouse(true)
	pullout:Hide()
	
	-- Scrollchild
	local scrollchild = CreateFrame("Frame", "UIF-StatusbarScrollChild"..count, pullout)
	
	-- Scrollframe
	local scrollframe = CreateFrame("ScrollFrame", "UIF-StatusbarScrollFrame"..count, pullout)
	scrollframe:SetPoint("TOPLEFT", 4, -2 - 5)
	scrollframe:SetPoint("BOTTOMRIGHT", -2, 2 + 5)
	scrollframe:SetScrollChild(scrollchild)
	scrollframe:SetScript("OnSizeChanged", ScrollFrame_OnSizeChanged)
	scrollframe:SetScript("OnMouseWheel", ScrollFrame_OnMouseWheel)
	
	-- Scrollbar
	local scrollbar = CreateFrame("Slider", "UIF-StatusbarScrollBar"..count, scrollframe, "UIPanelScrollBarTemplate")
	scrollbar:SetPoint("TOPRIGHT", scrollframe, -5, -18)
	scrollbar:SetPoint("BOTTOMRIGHT", scrollframe, -5, 18)
	scrollbar:SetMinMaxValues(0, 1000)
	scrollbar:SetValueStep(1)
	scrollbar:SetValue(0)
	scrollbar:SetWidth(10)
	scrollbar:SetScript("OnValueChanged", Scrollbar_OnValueChanged)
		
	local scrollbg = scrollbar:CreateTexture(nil, "BACKGROUND")
	scrollbg:SetAllPoints(scrollbar)
	scrollbg:SetTexture(0.4, 0.4, 0.4, 1)
	
	local test = pullout:CreateTexture()
	pullout.test = test
	
	local widget = {
		StatusbarFrames = {},
		frame = frame,
		type = Type,
		title = title,
		pane = pane,
		statusbar = statusbar,
		pullout = pullout,
		scrollframe = scrollframe,
		scrollchild = scrollchild,
		scrollbar = scrollbar,
	}
	
	for method, func in pairs(methods) do
		widget[method] = func
	end
	
	for i, v in pairs({scrollbar, scrollframe, button, pullout}) do
		v.obj = widget
	end
	
	return AceGUI:RegisterAsWidget(widget)
end
AceGUI:RegisterWidgetType(Type, Constructor, Version)